"""Unit test utilities."""
import filecmp
import pathlib


# Directory containing unit test input files, etc.
TESTDATA_DIR = pathlib.Path(__file__).parent/"testdata"


def assert_dirs_eq(dir1, dir2):
    """Compare two directories of files."""
    assert dir1 != dir2, (
        "Refusing to compare a directory to itself:\n"
        f"dir1 = {dir1}\n"
        f"dir2 = {dir2}\n"
    )

    # Get a list of files in each directory
    dir1 = pathlib.Path(dir1)
    dir2 = pathlib.Path(dir2)
    paths1 = list(dir1.iterdir())
    paths2 = list(dir2.iterdir())

    # Sanity checks
    assert paths1, f"Empty directory: {dir1}"
    assert paths2, f"Empty directory: {dir2}"
    assert all(p.is_file() for p in paths1)
    assert all(p.is_file() for p in paths2)
    assert len(paths1) == len(paths2), (
        "Number of output files does not match:\n"
        f"dir1 = {dir1}\n"
        f"dir2 = {dir2}\n"
        f"number of files in dir1 = {len(paths1)}\n"
        f"number of files in dir2 = {len(paths2)}\n"
    )

    # Compare files pairwise
    for path1, path2 in zip(sorted(paths1), sorted(paths2)):
        assert filecmp.cmp(path1, path2, shallow=False), (
            "Files do not match:\n"
            f"path1 = {path1}\n"
            f"path2 = {path2}\n"
        )



"""A light weight MapReduce framework for education.

Andrew DeOrio <awdeorio@umich.edu>

"""
import argparse
import logging
import pathlib
import shutil
import sys
import textwrap
import pkg_resources
from .mapreduce import mapreduce
from .exceptions import MadoopError


def main():
    """Parse command line arguments and options then call mapreduce()."""
    parser = argparse.ArgumentParser(
        description='A light weight MapReduce framework for education.'
    )

    optional_args = parser.add_argument_group('optional arguments')
    version = pkg_resources.get_distribution("madoop").version
    optional_args.add_argument(
        '--version', action='version',
        version=f'Madoop {version}'
    )
    optional_args.add_argument(
        '--example', action=ExampleAction, nargs=0,
        help="create example MapReduce program and input files",
    )
    optional_args.add_argument(
        '-v', '--verbose', action='count', default=0,
        help="verbose output"
    )
    required_args = parser.add_argument_group('required arguments')
    required_args.add_argument('-input', dest='input', required=True)
    required_args.add_argument('-output', dest='output', required=True)
    required_args.add_argument('-mapper', dest='mapper', required=True)
    required_args.add_argument('-reducer', dest='reducer', required=True)

    args, _ = parser.parse_known_args()

    # Handle verbose flag with logging configuration
    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('%(levelname)s: %(message)s')
    handler.setFormatter(formatter)
    root_logger = logging.getLogger()
    root_logger.addHandler(handler)
    if args.verbose > 0:
        root_logger.setLevel(logging.DEBUG)
    else:
        root_logger.setLevel(logging.INFO)

    # Run MapReduce API
    try:
        mapreduce(
            input_dir=args.input,
            output_dir=args.output,
            map_exe=args.mapper,
            reduce_exe=args.reducer,
        )
    except MadoopError as err:
        sys.exit(f"Error: {err}")


class ExampleAction(argparse.Action):
    """Copy example MapReduce program to PWD.

    We're using a custom Action because it runs before the check for required
    arguments executes.

    Doc: https://docs.python.org/3/library/argparse.html#argparse.Action
    """

    # Python 3.6 pylint bug work around
    # pylint: disable=too-few-public-methods

    def __init__(self, *args, **kwargs):
        """Call parent class init."""
        super().__init__(*args, **kwargs)

    def __call__(self, parser, *args, **kwargs):
        """Copy example/ directory to PWD."""
        madoop_dir = pathlib.Path(__file__).parent
        src = madoop_dir/"example"
        dst = pathlib.Path("example")
        if dst.exists():
            parser.error(f"directory already exists: {dst}")
        shutil.copytree(src, dst)
        print(textwrap.dedent(f"""\
            Created {dst}, try:

            madoop \\
              -input example/input \\
              -output example/output \\
              -mapper example/map.py \\
              -reducer example/reduce.py\
        """))
        parser.exit()


if __name__ == '__main__':
    main()



"""System tests for the API interface."""
import pytest
import madoop
from . import utils
from .utils import TESTDATA_DIR


def test_simple(tmpdir):
    """Run a simple MapReduce job and verify the output."""
    with tmpdir.as_cwd():
        madoop.mapreduce(
            input_dir=TESTDATA_DIR/"word_count/input",
            output_dir="output",
            map_exe=TESTDATA_DIR/"word_count/map.py",
            reduce_exe=TESTDATA_DIR/"word_count/reduce.py",
        )
    utils.assert_dirs_eq(
        TESTDATA_DIR/"word_count/correct/output",
        tmpdir/"output",
    )


def test_bash_executable(tmpdir):
    """Run a MapReduce job written in Bash."""
    with tmpdir.as_cwd():
        madoop.mapreduce(
            input_dir=TESTDATA_DIR/"word_count/input",
            output_dir="output",
            map_exe=TESTDATA_DIR/"word_count/map.sh",
            reduce_exe=TESTDATA_DIR/"word_count/reduce.sh",
        )
    utils.assert_dirs_eq(
        TESTDATA_DIR/"word_count/correct/output",
        tmpdir/"output",
    )


def test_bad_map_exe(tmpdir):
    """Map exe returns non-zero should produce an error message."""
    with tmpdir.as_cwd(), pytest.raises(madoop.MadoopError):
        madoop.mapreduce(
            input_dir=TESTDATA_DIR/"word_count/input",
            output_dir="output",
            map_exe=TESTDATA_DIR/"word_count/map_invalid.py",
            reduce_exe=TESTDATA_DIR/"word_count/reduce.py",
        )


def test_missing_shebang(tmpdir):
    """Reduce exe with a bad shebag should produce an error message."""
    with tmpdir.as_cwd(), pytest.raises(madoop.MadoopError):
        madoop.mapreduce(
            input_dir=TESTDATA_DIR/"word_count/input",
            output_dir="output",
            map_exe=TESTDATA_DIR/"word_count/map.py",
            reduce_exe=TESTDATA_DIR/"word_count/reduce_invalid.py",
        )


def test_empty_inputs(tmpdir):
    """Empty input files should not raise an error."""
    with tmpdir.as_cwd():
        madoop.mapreduce(
            input_dir=TESTDATA_DIR/"word_count/input_empty",
            output_dir="output",
            map_exe=TESTDATA_DIR/"word_count/map.py",
            reduce_exe=TESTDATA_DIR/"word_count/reduce.py",
        )
    utils.assert_dirs_eq(
        TESTDATA_DIR/"word_count/correct/output",
        tmpdir/"output",
    )



"""System tests for the command line interface."""
import subprocess
import pkg_resources
import pytest
from . import utils
from .utils import TESTDATA_DIR


def test_version():
    """Verify --version flag."""
    result = subprocess.run(
        ["madoop", "--version"],
        stdout=subprocess.PIPE,
        check=True,
    )
    output = result.stdout.decode("utf-8")
    assert "Madoop" in output
    assert pkg_resources.get_distribution("madoop").version in output


def test_help():
    """Verify --help flag."""
    result = subprocess.run(
        ["madoop", "--help"],
        stdout=subprocess.PIPE,
        check=True,
    )
    output = result.stdout.decode("utf-8")
    assert "usage" in output


def test_simple(tmpdir):
    """Run a simple MapReduce job and verify the output."""
    with tmpdir.as_cwd():
        subprocess.run(
            [
                "madoop",
                "-input", TESTDATA_DIR/"word_count/input",
                "-output", "output",
                "-mapper", TESTDATA_DIR/"word_count/map.py",
                "-reducer", TESTDATA_DIR/"word_count/reduce.py",
            ],
            stdout=subprocess.PIPE,
            check=True,
        )
    utils.assert_dirs_eq(
        TESTDATA_DIR/"word_count/correct/output",
        tmpdir/"output",
    )


def test_verbose(tmpdir):
    """Run a simple MapReduce job and verify the verbose stdout."""
    with tmpdir.as_cwd():
        completed_process = subprocess.run(
            [
                "madoop",
                "--verbose",
                "-input", TESTDATA_DIR/"word_count/input",
                "-output", "output",
                "-mapper", TESTDATA_DIR/"word_count/map.py",
                "-reducer", TESTDATA_DIR/"word_count/reduce.py",
            ],
            stdout=subprocess.PIPE,
            universal_newlines=True,
            check=True,
        )
    stdout_lines = completed_process.stdout.strip().split("\n")
    any(i.startswith("INFO") for i in stdout_lines)
    any(i.startswith("DEBUG") for i in stdout_lines)
    assert len(stdout_lines) > 20


def test_hadoop_arguments(tmpdir):
    """Hadoop Streaming arguments should be ignored."""
    with tmpdir.as_cwd():
        subprocess.run(
            [
                "madoop",
                "jar", "hadoop-streaming-2.7.2.jar",  # Hadoop args
                "-input", TESTDATA_DIR/"word_count/input",
                "-output", "output",
                "-mapper", TESTDATA_DIR/"word_count/map.py",
                "-reducer", TESTDATA_DIR/"word_count/reduce.py",
            ],
            stdout=subprocess.PIPE,
            check=True,
        )


def test_example(tmpdir):
    """Example option should copy files."""
    with tmpdir.as_cwd():
        subprocess.run(
            ["madoop", "--example"],
            check=True,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )
    assert (tmpdir/"example/input/input01.txt").exists()
    assert (tmpdir/"example/input/input02.txt").exists()
    assert (tmpdir/"example/map.py").exists()
    assert (tmpdir/"example/reduce.py").exists()

    # Call it again and it should refuse to clobber
    with tmpdir.as_cwd(), pytest.raises(subprocess.CalledProcessError):
        subprocess.run(
            ["madoop", "--example"],
            check=True,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )



"""System tests for the map stage of Michigan Hadoop."""
from pathlib import Path
from madoop.mapreduce import map_stage, group_stage, reduce_stage
from . import utils
from .utils import TESTDATA_DIR


def test_map_stage(tmpdir):
    """Test the map stage using word count example."""
    map_stage(
        exe=TESTDATA_DIR/"word_count/map.py",
        input_dir=TESTDATA_DIR/"word_count/input",
        output_dir=Path(tmpdir),
    )
    utils.assert_dirs_eq(
        TESTDATA_DIR/"word_count/correct/mapper-output",
        tmpdir,
    )


def test_group_stage(tmpdir):
    """Test group stage using word count example."""
    group_stage(
        input_dir=TESTDATA_DIR/"word_count/correct/mapper-output",
        output_dir=Path(tmpdir),
    )
    utils.assert_dirs_eq(
        TESTDATA_DIR/"word_count/correct/grouper-output",
        tmpdir,
    )


def test_reduce_stage(tmpdir):
    """Test reduce stage using word count example."""
    reduce_stage(
        exe=TESTDATA_DIR/"word_count/reduce.py",
        input_dir=TESTDATA_DIR/"word_count/correct/grouper-output",
        output_dir=Path(tmpdir),
    )
    utils.assert_dirs_eq(
        TESTDATA_DIR/"word_count/correct/reducer-output",
        tmpdir,
    )



"""Madoop build and install configuration."""
import pathlib
import setuptools


# Read the contents of README file
PROJECT_DIR = pathlib.Path(__file__).parent
README = PROJECT_DIR/"README.md"
LONG_DESCRIPTION = README.open(encoding="utf8").read()


setuptools.setup(
    name="madoop",
    description="A light weight MapReduce framework for education.",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    version="0.4.0",
    author="Andrew DeOrio",
    author_email="awdeorio@umich.edu",
    url="https://github.com/eecs485staff/madoop/",
    license="MIT",
    packages=["madoop"],
    keywords=[
        "madoop", "Hadoop", "MapReduce", "Michigan Hadoop", "Hadoop Streaming"
    ],
    install_requires=[],
    extras_require={
        "dev": [
            "pdbpp",
            "twine",
            "tox",
            "check-manifest",
            "freezegun",
            "pycodestyle",
            "pydocstyle",
            "pylint",
            "pytest",
            "pytest-cov",
        ],
    },
    python_requires='>=3.6',
    entry_points={
        "console_scripts": [
            "madoop = madoop.__main__:main",
        ]
    },
)



"""A light weight MapReduce framework for education.

Andrew DeOrio <awdeorio@umich.edu>

"""
import contextlib
import collections
import hashlib
import logging
import math
import pathlib
import shutil
import subprocess
import tempfile
from .exceptions import MadoopError


# Large input files are automatically split
MAX_INPUT_SPLIT_SIZE = 2**20  # 1 MB

# The number of reducers is dynamically determined by the number of unique keys
# but will not be more than MAX_NUM_REDUCE
MAX_NUM_REDUCE = 4

# Madoop logger
LOGGER = logging.getLogger("madoop")


def mapreduce(input_dir, output_dir, map_exe, reduce_exe):
    """Madoop API."""
    # Do not clobber existing output directory
    output_dir = pathlib.Path(output_dir)
    if output_dir.exists():
        raise MadoopError(f"Output directory already exists: {output_dir}")
    output_dir.mkdir()

    # Executable scripts must have valid shebangs
    is_executable(map_exe)
    is_executable(reduce_exe)

    # Create a tmp directory which will be automatically cleaned up
    with tempfile.TemporaryDirectory(prefix="madoop-") as tmpdir:
        tmpdir = pathlib.Path(tmpdir)
        LOGGER.debug("tmpdir=%s", tmpdir)

        # Create stage input and output directory
        map_input_dir = tmpdir/'input'
        map_output_dir = tmpdir/'mapper-output'
        reduce_input_dir = tmpdir/'reducer-input'
        reduce_output_dir = tmpdir/'output'
        map_input_dir.mkdir()
        map_output_dir.mkdir()
        reduce_input_dir.mkdir()
        reduce_output_dir.mkdir()

        # Copy and rename input files: part-00000, part-00001, etc.
        input_dir = pathlib.Path(input_dir)
        prepare_input_files(input_dir, map_input_dir)

        # Executables must be absolute paths
        map_exe = pathlib.Path(map_exe).resolve()
        reduce_exe = pathlib.Path(reduce_exe).resolve()

        # Run the mapping stage
        LOGGER.info("Starting map stage")
        map_stage(
            exe=map_exe,
            input_dir=map_input_dir,
            output_dir=map_output_dir,
        )

        # Run the grouping stage
        LOGGER.info("Starting group stage")
        group_stage(
            input_dir=map_output_dir,
            output_dir=reduce_input_dir,
        )

        # Run the reducing stage
        LOGGER.info("Starting reduce stage")
        reduce_stage(
            exe=reduce_exe,
            input_dir=reduce_input_dir,
            output_dir=reduce_output_dir,
        )

        # Move files from temporary output dir to user-specified output dir
        total_size = 0
        for filename in sorted(reduce_output_dir.glob("*")):
            st_size = filename.stat().st_size
            total_size += st_size
            shutil.copy(filename, output_dir)
            output_path = output_dir.parent/last_two(filename)
            LOGGER.debug("%s size=%sB", output_path, st_size)

    # Remind user where to find output
    LOGGER.debug("total output size=%sB", total_size)
    LOGGER.info("Output directory: %s", output_dir)


def prepare_input_files(input_dir, output_dir):
    """Copy and split input files.  Rename to part-00000, part-00001, etc.

    If a file in input_dir is smaller than MAX_INPUT_SPLIT_SIZE, then copy it
    to output_dir.  For larger files, split into blocks of MAX_INPUT_SPLIT_SIZE
    bytes and write block to output_dir. Input files will never be combined.

    The number of files created will be the number of mappers since we will
    assume that the number of tasks per mapper is 1.  Apache Hadoop has a
    configurable number of tasks per mapper, however for both simplicity and
    because our use case has smaller inputs we use 1.

    """
    assert input_dir.is_dir(), f"Can't find input_dir '{input_dir}'"

    # Split and copy input files
    part_num = 0
    total_size = 0
    for inpath in sorted(input_dir.glob('*')):
        assert inpath.is_file()

        # Compute output filenames
        st_size = inpath.stat().st_size
        total_size += st_size
        n_splits = math.ceil(st_size / MAX_INPUT_SPLIT_SIZE)
        n_splits = 1 if not n_splits else n_splits  # Handle empty input file
        LOGGER.debug(
            "input %s size=%sB partitions=%s", inpath, st_size, n_splits
        )
        outpaths = [
            output_dir/part_filename(part_num + i) for i in range(n_splits)
        ]
        part_num += n_splits

        # Copy to new output files
        with contextlib.ExitStack() as stack:
            outfiles = [stack.enter_context(i.open('w')) for i in outpaths]
            infile = stack.enter_context(inpath.open(encoding="utf-8"))
            outparent = outpaths[0].parent
            assert all(i.parent == outparent for i in outpaths)
            outnames = [i.name for i in outpaths]
            logging.debug(
                "partition %s >> %s/{%s}",
                last_two(inpath), outparent.name, ",".join(outnames),
            )
            for i, line in enumerate(infile):
                outfiles[i % n_splits].write(line)
    LOGGER.debug("total input size=%sB", total_size)


def is_executable(exe):
    """Verify exe is executable and raise exception if it is not.

    Execute exe with an empty string input and verify that it returns zero.  We
    can't just check the executable bit because scripts with incorrect shebangs
    result in difficult-to-understand error messages.

    """
    try:
        subprocess.run(
            str(exe),
            shell=True,
            input="".encode(),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=True,
        )
    except subprocess.CalledProcessError as err:
        raise MadoopError(f"Failed executable test: {err}") from err


def part_filename(num):
    """Return a string conforming to the output filename convention.

    EXAMPLE
    >>> part_filename(3)
    'part-00003'

    """
    return f"part-{num:05d}"


def map_stage(exe, input_dir, output_dir):
    """Execute mappers."""
    i = 0
    for i, input_path in enumerate(sorted(input_dir.iterdir()), 1):
        output_path = output_dir/part_filename(i)
        LOGGER.debug(
            "%s < %s > %s",
            exe.name, last_two(input_path), last_two(output_path),
        )
        with input_path.open() as infile, output_path.open('w') as outfile:
            try:
                subprocess.run(
                    str(exe),
                    shell=True,
                    check=True,
                    stdin=infile,
                    stdout=outfile,
                )
            except subprocess.CalledProcessError as err:
                raise MadoopError(
                    f"Command returned non-zero: "
                    f"{exe} < {input_path} > {output_path}"
                ) from err
    LOGGER.info("Finished map executions: %s", i)


def sort_file(path):
    """Sort contents of path, overwriting it."""
    LOGGER.debug("sort %s", last_two(path))
    with path.open() as infile:
        sorted_lines = sorted(infile)
    with path.open("w") as outfile:
        outfile.writelines(sorted_lines)


def keyhash(key):
    """Hash key and return an integer."""
    hexdigest = hashlib.md5(key.encode("utf-8")).hexdigest()
    return int(hexdigest, base=16)


def partition_keys(inpath, outpaths, input_keys_stats, output_keys_stats):
    """Allocate lines of inpath among outpaths using hash of key.

    Update the data structures provided by the caller input_keys_stats and
    output_keys_stats.  Both map a filename to a set of of keys.

    """
    assert len(outpaths) == MAX_NUM_REDUCE
    outparent = outpaths[0].parent
    assert all(i.parent == outparent for i in outpaths)
    with contextlib.ExitStack() as stack:
        outfiles = [stack.enter_context(p.open("a")) for p in outpaths]
        for line in stack.enter_context(inpath.open()):
            key = line.partition('\t')[0]
            input_keys_stats[inpath].add(key)
            reducer_idx = keyhash(key) % MAX_NUM_REDUCE
            outfiles[reducer_idx].write(line)
            outpath = outpaths[reducer_idx]
            output_keys_stats[outpath].add(key)


def group_stage(input_dir, output_dir):
    """Run group stage.

    Process each mapper output file, allocating lines to grouper output files
    using the hash and modulo of the key.

    """
    # Compute output filenames
    outpaths = []
    for i in range(MAX_NUM_REDUCE):
        outpaths.append(output_dir/part_filename(i))

    # Track keyspace stats, map filename -> set of keys
    input_keys_stats = collections.defaultdict(set)
    output_keys_stats = collections.defaultdict(set)

    # Partition input, appending to output files
    for inpath in sorted(input_dir.iterdir()):
        partition_keys(inpath, outpaths, input_keys_stats, output_keys_stats)

    # Log input keyspace stats
    all_input_keys = set()
    for inpath, keys in sorted(input_keys_stats.items()):
        all_input_keys.update(keys)
        LOGGER.debug("%s unique_keys=%s", last_two(inpath), len(keys))
    LOGGER.debug("%s all_unique_keys=%s", input_dir.name, len(all_input_keys))

    # Log partition input and output filenames
    outnames = [i.name for i in outpaths]
    outparent = outpaths[0].parent
    for inpath in sorted(input_keys_stats.keys()):
        LOGGER.debug(
            "partition %s >> %s/{%s}",
            last_two(inpath), outparent.name, ",".join(outnames),
        )

    # Sort output files
    for path in sorted(output_dir.iterdir()):
        sort_file(path)

    # Log output keyspace stats
    all_output_keys = set()
    for outpath, keys in sorted(output_keys_stats.items()):
        all_output_keys.update(keys)
        LOGGER.debug("%s unique_keys=%s", last_two(outpath), len(keys))
    LOGGER.debug("%s all_unique_keys=%s", output_dir.name,
                 len(all_output_keys))


def reduce_stage(exe, input_dir, output_dir):
    """Execute reducers."""
    i = 0
    for i, input_path in enumerate(sorted(input_dir.iterdir())):
        output_path = output_dir/part_filename(i)
        LOGGER.debug(
            "%s < %s > %s",
            exe.name, last_two(input_path), last_two(output_path),
        )
        with input_path.open() as infile, output_path.open('w') as outfile:
            try:
                subprocess.run(
                    str(exe),
                    shell=True,
                    check=True,
                    stdin=infile,
                    stdout=outfile,
                )
            except subprocess.CalledProcessError as err:
                raise MadoopError(
                    f"Command returned non-zero: "
                    f"{exe} < {input_path} > {output_path}"
                ) from err
    LOGGER.info("Finished reduce executions: %s", i+1)


def last_two(path):
    """Return the last two parts of path."""
    return pathlib.Path(*path.parts[-2:])



#!/usr/bin/env python3
"""Word count reducer."""
import sys
import itertools


def main():
    """Divide sorted lines into groups that share a key."""
    for key, group in itertools.groupby(sys.stdin, keyfunc):
        reduce_one_group(key, group)


def keyfunc(line):
    """Return the key from a TAB-delimited key-value pair."""
    return line.partition("\t")[0]


def reduce_one_group(key, group):
    """Reduce one group."""
    word_count = 0
    for line in group:
        count = line.partition("\t")[2]
        word_count += int(count)
    print(f"{key} {word_count}")


if __name__ == "__main__":
    main()
